__version__ = "1.0.1"

from .functional import dic_del, dic_get, dic_has, dic_merge, dic_set, dic_walk
from .wrapper import DictBelt

__all__ = [
    "dic_del",
    "dic_get",
    "dic_has",
    "dic_merge",
    "dic_set",
    "dic_walk",
    "DictBelt",
]



import copy
from functools import lru_cache
from typing import Any, Dict, List

DictKey = List[Any]


def dic_get(
    dic: Dict, key: DictKey, *, default: Any = None, raise_on_missing: bool = True
) -> Any:
    """
    Retrieve the value at `key` in `dic`.

    `default` is the default value to return if missing
    `raise_on_missing` indicates whether to raise an exception if the key is not found
    """
    buffer = dic
    for segment in key:
        dict_error = type(buffer) is dict and segment not in buffer
        list_error = type(buffer) is list and (
            type(segment) is not int or len(buffer) <= segment
        )

        if dict_error or list_error:
            if default is None and raise_on_missing:
                raise KeyError(f"'{'.'.join(map(str, key))}' not found in dict")
            else:
                return default

        buffer = buffer[segment]
    return buffer


def dic_set(dic: Dict, key: DictKey, val: Any) -> Dict:
    """
    Set the value `val` in `dic` at `key` and returns `dic`.

    Will raise an error if key doesn't exist.
    """
    if len(key) == 1:
        dic[key[0]] = val
        return
    parent = dic_get(dic, key[:-1])
    parent[key[-1]] = val
    return dic


def dic_del(dic: Dict, key: DictKey) -> Dict:
    """
    Delete `key` from dictionary `dic`. Mutates `dic`.
    """
    parent = dic_get(dic, key[:-1])
    parent[key[-1]] = None
    del parent[key[-1]]
    return dic


def dic_has(dic: Dict, key: DictKey = list()):
    """
    Predicate to verify if `key` exists in `dic`.
    """
    try:
        dic_get(dic, key)
    except KeyError:
        return False
    return True


def dic_walk(
    dic: Dict, key: DictKey = list(), callback: callable = None, context: Dict = dict()
) -> Dict:
    """
    Invokes `callback` for each item of `dic` including each sub-dict and sub-list thereof.

    The expected signature of `callback` is like so:

    ```
    def my_callback(dic: Dict, key: DictKey, val: Any, context: Dict)
    ```

    `context` is a dictionary passed along to each invocation of `callback`.
    """
    subject = dic_get(dic, key)

    iterator = None
    if type(subject) is dict:
        iterator = subject.items()
    elif type(subject) is list:
        iterator = enumerate(subject)
    else:
        return subject

    for segment, val in iterator:
        local_context = copy.deepcopy(context)
        if callback:
            try:
                ret = callback(dic, key + [segment], val, local_context)
                if type(ret) is dict:
                    local_context = ret
            except StopIteration:
                return
        if type(val) is dict or type(val) is list:
            dic_walk(dic, key=key + [segment], callback=callback, context=local_context)
    return dic


def dic_merge(dicts: List[Dict]) -> Dict:
    """
    Merge all `dicts` into one. Mutates `dicts[0]`.
    """

    def _merge(dic1, dic2):
        for k, v in dic2.items():
            if k in dic1 and isinstance(dic1[k], dict) and isinstance(dic2[k], dict):
                _merge(dic1[k], dic2[k])
            else:
                dic1[k] = dic2[k]

    result = {}
    for dic in dicts:
        _merge(result, dic)
    return result



from typing import Any, Dict, List, Optional, Union

from .functional import dic_del, dic_get, dic_has, dic_merge, dic_set, dic_walk

DictKey = Union[Any, List[Any]]


def normalise_key(key):
    return [key] if type(key) is not list else key


class DictBelt:
    """
    Class wrapper to access utility methods in an objective fashion.
    """

    def __init__(self, data: Dict):
        self.dict = data

    def get(
        self, key: DictKey, *, default: Any = None, raise_on_missing: bool = True
    ) -> Any:
        """
        Retrieve the value at `key` in `self.dict`.

        `default` is the default value to return if missing
        `raise_on_missing` indicates whether to raise an exception if the key is not found
        """
        value = dic_get(
            self.dict,
            normalise_key(key),
            default=default,
            raise_on_missing=raise_on_missing,
        )
        if type(value) is dict:
            return DictBelt(value)
        return value

    def set(self, key: DictKey, value: Any) -> "DictBelt":
        """
        Set the value `val` in `self.dict` at `key` and returns self.

        Will raise an error if key doesn't exist.
        """
        dic_set(self.dict, normalise_key(key), value)
        return self

    def rm(self, key: DictKey) -> "DictBelt":
        """
        Delete `key` from dictionary `self.dict`. Mutates `self.dict` and return `self`.
        """
        dic_del(self.dict, normalise_key(key))
        return self

    def has(self, key: DictKey) -> bool:
        """
        Predicate to verify if `key` exists in `self.dict`.
        """
        return dic_has(self.dict, normalise_key(key))

    def walk(
        self,
        key: DictKey = list(),
        callback: Optional[callable] = None,
        context: Dict = dict(),
    ) -> "DictBelt":
        """
        Invokes `callback` for each item of `self.dict` including each sub-dict and sub-list thereof.

        The expected signature of `callback` is like so:

        ```
        def my_callback(dic: Dict, key: DictKey, val: Any, context: Dict)
        ```

        `context` is a dictionary passed along to each invocation of `callback`.
        """
        dic_walk(self.dict, normalise_key(key), callback, context)
        return self

    def merge(self, *dicts):
        """
        Merge `self.dict` with all `dicts` into one. Mutates `self.dict`.
        """

        def normalise_dict(d):
            return d if type(d) is dict else d.dict

        return dic_merge(map(normalise_dict, [self, *dicts]))

    def __add__(self, b):
        return self.merge(b)

    def __getitem__(self, key: DictKey) -> Any:
        return self.get(key)

    def __setitem__(self, key: DictKey, value: Any):
        self.set(key, value)

    def __contains__(self, key: DictKey):
        return self.has(key)

    def __len__(self) -> int:
        return len(self.dict)

    def __delitem__(self, key: DictKey):
        self.rm(key)

    def keys(self):
        return self.dict.keys()

    def items(self):
        return self.dict.items()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dictbelt']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'dictbelt',
    'version': '1.0.1',
    'description': 'A batbelt utility for dictionaries',
    'long_description': 'DICTBELT\n========\n\nThe bat-belt for dictionaries!\n\nThis library contains a set of functions and wrapper to better deal with complex\nhierarchies of dictionaries (and lists).\n\nOf note is the `dic_walk()` function that allows traversing an entire `dict` and\n`list` tree.\n\n',
    'author': 'Steeleye',
    'author_email': 'python.devs@steel-eye.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)
